@@ -167,34 +167,26 @@ fn as_codepoint(v: &Val) -> Result<char, Error> {
     char::from_u32(u).ok_or_else(|| Error::str(format_args!("cannot use {u} as character")))
 }
 
-/// Split a string by a given separator string.
-fn split(s: &str, sep: &str) -> Vec<Val> {
-    if sep.is_empty() {
-        // Rust's `split` function with an empty separator ("")
-        // yields an empty string as first and last result
-        // to prevent this, we are using `chars` instead
-        s.chars().map(|s| Val::str(s.to_string())).collect()
-    } else {
-        s.split(sep).map(|s| Val::str(s.to_string())).collect()
-    }
-}
-
 /// This implements a ~10x faster version of:
 /// ~~~ text
 /// def range($from; $to; $by): $from |
-///    if $by > 0 then while(. < $to; . + $by)
-///    else            while(. > $to; . + $by)
+///    if $by > 0 then while(.  < $to; . + $by)
+///  elif $by < 0 then while(.  > $to; . + $by)
+///    else            while(. != $to; . + $by)
 ///    end;
 /// ~~~
 fn range(mut from: ValR, to: Val, by: Val) -> impl Iterator<Item = ValR> {
-    let positive = by > Val::Int(0);
+    let cmp = by.cmp(&Val::Int(0));
+    use core::cmp::Ordering::{Equal, Greater, Less};
     core::iter::from_fn(move || match from.clone() {
-        Ok(x) if (positive && x < to) || (!positive && x > to) => {
-            Some(core::mem::replace(&mut from, x + by.clone()))
+        Ok(x) => match cmp {
+            Greater => x < to,
+            Less => x > to,
+            Equal => x != to,
         }
-        Ok(_) => None,
+        .then(|| core::mem::replace(&mut from, x + by.clone())),
         e @ Err(_) => {
-            // return None as following value
+            // return None after the error
             from = Ok(to.clone());
             Some(e)
         }
@@ -232,6 +224,36 @@ fn to_csv(vs: &[Val]) -> Result<String, Error> {
     Ok(vs.iter().map(fr).collect::<Result<Vec<_>, _>>()?.join(","))
 }
 
+/// Return the string windows having `n` characters, where `n` > 0.
+///
+/// Taken from <https://users.rust-lang.org/t/iterator-over-windows-of-chars/17841/3>.
+fn str_windows(line: &str, n: usize) -> impl Iterator<Item = &str> {
+    line.char_indices()
+        .zip(line.char_indices().skip(n).chain(Some((line.len(), ' '))))
+        .map(move |((i, _), (j, _))| &line[i..j])
+}
+
+/// Return the indices of `y` in `x`.
+fn indices<'a>(x: &'a Val, y: &'a Val) -> Result<Box<dyn Iterator<Item = usize> + 'a>, Error> {
+    match (x, y) {
+        (Val::Str(_), Val::Str(y)) if y.is_empty() => Ok(Box::new(core::iter::empty())),
+        (Val::Arr(_), Val::Arr(y)) if y.is_empty() => Ok(Box::new(core::iter::empty())),
+        (Val::Str(x), Val::Str(y)) => {
+            let iw = str_windows(x, y.chars().count()).enumerate();
+            Ok(Box::new(iw.filter_map(|(i, w)| (w == **y).then_some(i))))
+        }
+        (Val::Arr(x), Val::Arr(y)) => {
+            let iw = x.windows(y.len()).enumerate();
+            Ok(Box::new(iw.filter_map(|(i, w)| (w == **y).then_some(i))))
+        }
+        (Val::Arr(x), y) => {
+            let ix = x.iter().enumerate();
+            Ok(Box::new(ix.filter_map(move |(i, x)| (x == y).then_some(i))))
+        }
+        (x, y) => Err(Error::Index(x.clone(), y.clone())),
+    }
+}
+
 fn once_with<'a, T>(f: impl FnOnce() -> T + 'a) -> Box<dyn Iterator<Item = T> + 'a> {
     Box::new(core::iter::once_with(f))
 }
@@ -304,9 +326,11 @@ const CORE_RUN: &[(&str, usize, RunPtr)] = &[
         let vals = args.get(0).run(cv.clone());
         Box::new(vals.map(move |y| Ok(Val::Bool(cv.1.contains(&y?)))))
     }),
-    ("split", 1, |args, cv| {
-        let seps = args.get(0).run(cv.clone());
-        Box::new(seps.map(move |sep| Ok(Val::arr(split(cv.1.as_str()?, sep?.as_str()?)))))
+    ("indices", 1, |args, cv| {
+        let vals = args.get(0).run(cv.clone());
+        let to_int = |i: usize| Val::Int(i.try_into().unwrap());
+        let f = move |v| indices(&cv.1, &v?).map(|idxs| Val::arr(idxs.map(to_int).collect()));
+        Box::new(vals.map(f))
     }),
     ("first", 1, |args, cv| Box::new(args.get(0).run(cv).take(1))),
     ("limit", 2, |args, cv| {
