@@ -157,6 +157,27 @@ fn has() {
     give(json!({"a": 1, "b": null}), r#"has("c")"#, json!(false));
 }
 
+yields!(indices_str, r#""a,b, cd, efg" | indices(", ")"#, [3, 7]);
+yields!(
+    indices_arr_num,
+    "[0, 1, 2, 1, 3, 1, 4] | indices(1)",
+    [1, 3, 5]
+);
+yields!(
+    indices_arr_arr,
+    "[0, 1, 2, 3, 1, 4, 2, 5, 1, 2, 6, 7] | indices([1, 2])",
+    [1, 8]
+);
+yields!(indices_arr_str, r#"["a", "b", "c"] | indices("b")"#, [1]);
+
+yields!(indices_arr_empty, "[0, 1] | indices([])", json!([]));
+yields!(indices_arr_larger, "[1, 2] | indices([1, 2, 3])", json!([]));
+
+yields!(indices_arr_overlap, "[0, 0, 0] | indices([0, 0])", [0, 1]);
+yields!(indices_str_overlap, r#""aaa" | indices("aa")"#, [0, 1]);
+yields!(indices_str_gb1, r#""🇬🇧!" | indices("!")"#, [2]);
+yields!(indices_str_gb2, r#""🇬🇧🇬🇧" | indices("🇬🇧")"#, [0, 2]);
+
 #[test]
 fn json() {
     // TODO: correct this
@@ -177,23 +198,17 @@ fn keys_unsorted() {
     fail(json!(null), "keys_unsorted", err(Val::Null));
 }
 
-#[test]
-fn length() {
-    give(json!("ƒoo"), "length", json!(3));
-    give(json!("नमस्ते"), "length", json!(6));
-    give(json!({"a": 5, "b": 3}), "length", json!(2));
-    give(json!(2), "length", json!(2));
-    give(json!(-2), "length", json!(2));
-    give(json!(2.5), "length", json!(2.5));
-    give(json!(-2.5), "length", json!(2.5));
-}
+yields!(length_str_foo, r#""ƒoo" | length"#, 3);
+yields!(length_str_namaste, r#""नमस्ते" | length"#, 6);
+yields!(length_obj, r#"{"a": 5, "b": 3} | length"#, 2);
+yields!(length_int_pos, " 2 | length", 2);
+yields!(length_int_neg, "-2 | length", 2);
+yields!(length_float_pos, " 2.5 | length", 2.5);
+yields!(length_float_neg, "-2.5 | length", 2.5);
 
-#[test]
-fn utf8bytelength() {
-    give(json!("foo"), "utf8bytelength", json!(3));
-    give(json!("ƒoo"), "utf8bytelength", json!(4));
-    give(json!("नमस्ते"), "utf8bytelength", json!(18));
-}
+yields!(utf8bytelength_foo1, r#""foo" | utf8bytelength"#, 3);
+yields!(utf8bytelength_foo2, r#""ƒoo" | utf8bytelength"#, 4);
+yields!(utf8bytelength_namaste, r#""नमस्ते" | utf8bytelength"#, 18);
 
 #[test]
 fn limit() {
@@ -243,6 +258,18 @@ yields!(
     [10.0, 11.0, 12.0, 13.0, 7.0, 8.0, 8.0, 9.0]
 );
 
+yields!(range_pp, "[range(0; 6;  2)]", [0, 2, 4]);
+yields!(range_pn, "[range(0; 6; -2)]", json!([]));
+yields!(range_np, "[range(0; -6; 2)]", json!([]));
+yields!(range_nn, "[range(0; -6; -2)]", [0, -2, -4]);
+yields!(range_zz, "[range(0; 0; 0)]", json!([]));
+yields!(range_fp, "[range(0.0; 2; 0.5)]", [0.0, 0.5, 1.0, 1.5]);
+yields!(range_ip, "[limit(3; range(0; 1/0; 1))]", [0, 1, 2]);
+yields!(range_in, "[limit(3; range(0; -1/0; -1))]", [0, -1, -2]);
+// here, we diverge from jq, which just returns the empty list
+yields!(range_pz, "[limit(3; range(0; 6; 0))]", json!([0, 0, 0]));
+yields!(range_nz, "[limit(3; range(0; -6; 0))]", json!([0, 0, 0]));
+
 #[test]
 fn regex() {
     let date = r#"(\\d{4})-(\\d{2})-(\\d{2})"#;
@@ -292,16 +319,6 @@ fn round() {
     fail(json!({}), "round", err(json!({})));
 }
 
-#[test]
-fn split() {
-    give(json!("aöß"), r#"split("")"#, json!(["a", "ö", "ß"]));
-    give(
-        json!("abcabcdab"),
-        r#"split("ab")"#,
-        json!(["", "c", "cd", ""]),
-    );
-}
-
 #[test]
 fn startswith() {
     give(json!("foobar"), r#"startswith("")"#, json!(true));
