@@ -1,24 +1,40 @@
 use crate::box_iter::{box_once, flat_map_with, map_with, BoxIter};
 use crate::results::{fold, recurse, then, Fold, Results};
-use crate::val::{Val, ValR, ValRs};
+use crate::val::{Val, ValR2, ValR2s, ValT};
 use crate::{rc_lazy_list, Bind, Ctx, Error};
 use alloc::{boxed::Box, string::String, vec::Vec};
+use core::ops::ControlFlow;
 use dyn_clone::DynClone;
 use jaq_syn::filter::FoldType;
 use jaq_syn::{MathOp, OrdOp};
 
 /// Function from a value to a stream of value results.
 #[derive(Debug, Clone)]
-pub struct Owned(Id, Box<[Ast]>);
+pub struct Owned<V = Val>(Id, Lut<V>);
 
-impl Default for Owned {
+/// Look-up table for indices stored in ASTs.
+#[derive(Clone, Debug)]
+struct Lut<V = Val> {
+    defs: Box<[Ast]>,
+    natives: Box<[Native<V>]>,
+}
+
+impl<V> Default for Owned<V> {
     fn default() -> Self {
-        Self(Id(0), Box::new([Ast::Id]))
+        Self::new(Id(0), [Ast::Id].into(), [].into())
     }
 }
 
-#[derive(Debug, Copy, Clone)]
-pub struct Ref<'a>(Id, &'a [Ast]);
+#[derive(Debug)]
+pub struct Ref<'a, V = Val>(Id, &'a Lut<V>);
+
+impl<'a, V> Clone for Ref<'a, V> {
+    fn clone(&self) -> Self {
+        *self
+    }
+}
+
+impl<'a, V> Copy for Ref<'a, V> {}
 
 #[derive(Default, Debug, Copy, Clone, PartialEq, Eq)]
 pub struct Id(pub usize);
@@ -34,7 +50,7 @@ pub enum CallTyp {
 }
 
 #[derive(Clone, Debug, PartialEq, Eq)]
-pub struct TailCall(Id, crate::Vars, Val);
+pub struct TailCall<V>(Id, crate::Vars<V>, V);
 
 #[derive(Clone, Debug)]
 pub(crate) struct Call {
@@ -44,24 +60,24 @@ pub(crate) struct Call {
     pub args: Box<[Bind<Id, Id>]>,
 }
 
-impl Owned {
-    pub(crate) fn new(main: Id, recs: Vec<Ast>) -> Self {
-        Self(main, recs.into())
+impl<V> Owned<V> {
+    pub(crate) fn new(main: Id, defs: Box<[Ast]>, natives: Box<[Native<V>]>) -> Self {
+        Self(main, Lut { defs, natives })
     }
 }
 
-/// Function from a value to a stream of value results.
-#[derive(Clone, Debug, Default)]
+/// Abstract syntax tree for filters.
+#[derive(Clone, Debug)]
 pub(crate) enum Ast {
-    #[default]
     Id,
     ToString,
 
     Int(isize),
-    Float(f64),
+    Num(String),
     Str(String),
     Array(Id),
-    Object(Box<[(Id, Id)]>),
+    ObjEmpty,
+    ObjSingle(Id, Id),
 
     Try(Id, Id),
     Neg(Id),
@@ -107,25 +123,26 @@ pub(crate) enum Ast {
     Var(usize),
     Call(Call),
 
-    Native(Native, Box<[Id]>),
+    Native(usize, Box<[Id]>),
 }
 
 // we can unfortunately not make a `Box<dyn ... + Clone>`
 // that is why we have to go through the pain of making a new trait here
-pub trait Update<'a>: Fn(Val) -> ValRs<'a> + DynClone {}
+pub trait Update<'a, V = Val>: Fn(V) -> ValR2s<'a, V> + DynClone {}
 
-impl<'a, T: Fn(Val) -> ValRs<'a> + Clone> Update<'a> for T {}
+impl<'a, V, T: Fn(V) -> ValR2s<'a, V> + Clone> Update<'a, V> for T {}
 
-dyn_clone::clone_trait_object!(<'a> Update<'a>);
+dyn_clone::clone_trait_object!(<'a, V> Update<'a, V>);
 
 /// Enhance the context `ctx` with variables bound to the outputs of `args` executed on `cv`,
 /// and return the enhanced contexts together with the original value of `cv`.
 ///
 /// This is used when we call filters with variable arguments.
-fn bind_vars<'a, I>(mut args: I, ctx: Ctx<'a>, cv: Cv<'a>) -> Results<'a, Cv<'a>, Error>
-where
-    I: Iterator<Item = Bind<Ref<'a>, Ref<'a>>> + Clone + 'a,
-{
+fn bind_vars<'a, V: ValT>(
+    mut args: impl Iterator<Item = Bind<Ref<'a, V>, Ref<'a, V>>> + Clone + 'a,
+    ctx: Ctx<'a, V>,
+    cv: Cv<'a, V>,
+) -> Results<'a, Cv<'a, V>, Error<V>> {
     match args.next() {
         Some(Bind::Var(arg)) => flat_map_with(
             arg.run(cv.clone()),
@@ -137,117 +154,108 @@ where
     }
 }
 
-fn run_cvs<'a>(f: Ref<'a>, cvs: Results<'a, Cv<'a>, Error>) -> impl Iterator<Item = ValR> + 'a {
-    cvs.flat_map(move |cv| then(cv, |cv| f.run(cv)))
+fn run_cvs<'a, V: ValT>(f: Ref<'a, V>, cvs: Results<'a, Cv<'a, V>, Error<V>>) -> ValR2s<'a, V> {
+    Box::new(cvs.flat_map(move |cv| then(cv, |cv| f.run(cv))))
 }
 
-type ObjVec = Vec<(ValR, ValR)>;
-fn obj_cart<'a, I>(mut args: I, cv: Cv<'a>, prev: ObjVec) -> BoxIter<'a, ObjVec>
+fn reduce<'a, T, V, F>(xs: Results<'a, T, Error<V>>, init: V, f: F) -> ValR2s<V>
 where
-    I: Iterator<Item = (Ref<'a>, Ref<'a>)> + Clone + 'a,
-{
-    if let Some((l, r)) = args.next() {
-        let iter = l.run(cv.clone());
-        flat_map_with(iter, (args, cv, prev), move |l, (args, cv, prev)| {
-            let iter = r.run(cv.clone());
-            flat_map_with(iter, (l, args, cv, prev), |r, (l, args, cv, mut prev)| {
-                prev.push((l, r));
-                obj_cart(args, cv, prev)
-            })
-        })
-    } else {
-        box_once(prev)
-    }
-}
-
-fn reduce<'a, T: Clone + 'a, F>(xs: Results<'a, T, Error>, init: Val, f: F) -> ValRs
-where
-    F: Fn(T, Val) -> ValRs<'a> + 'a,
+    T: Clone + 'a,
+    V: Clone + 'a,
+    F: Fn(T, V) -> ValR2s<'a, V> + 'a,
 {
     let xs = rc_lazy_list::List::from_iter(xs);
     Box::new(fold(false, xs, Fold::Input(init), f))
 }
 
-type Cv<'c> = (Ctx<'c>, Val);
+type Cv<'c, V = Val> = (Ctx<'c, V>, V);
 
 /// A filter which is implemented using function pointers.
 #[derive(Clone)]
-pub struct Native {
-    run: RunPtr,
-    update: UpdatePtr,
+pub struct Native<V = Val> {
+    run: RunPtr<V>,
+    update: UpdatePtr<V>,
 }
 
 /// Run function pointer.
-pub type RunPtr = for<'a> fn(Args<'a>, Cv<'a>) -> ValRs<'a>;
+pub type RunPtr<V = Val> = for<'a> fn(Args<'a, V>, Cv<'a, V>) -> ValR2s<'a, V>;
 /// Update function pointer.
-pub type UpdatePtr = for<'a> fn(Args<'a>, Cv<'a>, Box<dyn Update<'a> + 'a>) -> ValRs<'a>;
+pub type UpdatePtr<V = Val> =
+    for<'a> fn(Args<'a, V>, Cv<'a, V>, Box<dyn Update<'a, V> + 'a>) -> ValR2s<'a, V>;
 
-impl Native {
+impl<V> Native<V> {
     /// Create a native filter from a run function, without support for updates.
-    pub const fn new(run: RunPtr) -> Self {
+    pub const fn new(run: RunPtr<V>) -> Self {
         Self::with_update(run, |_, _, _| box_once(Err(Error::PathExp)))
     }
 
     /// Create a native filter from a run function and an update function (used for `filter |= ...`).
-    pub const fn with_update(run: RunPtr, update: UpdatePtr) -> Self {
+    // TODO for v2.0: remove this
+    pub const fn with_update(run: RunPtr<V>, update: UpdatePtr<V>) -> Self {
         Self { run, update }
     }
 }
 
-impl core::fmt::Debug for Native {
+// TODO for v2.0: remove this
+impl<V> core::fmt::Debug for Native<V> {
     fn fmt(&self, f: &mut core::fmt::Formatter) -> core::fmt::Result {
         f.debug_struct("Native").finish()
     }
 }
 
 /// Arguments passed to a native filter.
-#[derive(Copy, Clone)]
-pub struct Args<'a>(&'a [Id], &'a [Ast]);
+pub struct Args<'a, V = Val>(&'a [Id], &'a Lut<V>);
 
-impl<'a> Args<'a> {
+impl<'a, V> Clone for Args<'a, V> {
+    fn clone(&self) -> Self {
+        *self
+    }
+}
+
+impl<'a, V> Copy for Args<'a, V> {}
+
+impl<'a, V: ValT> Args<'a, V> {
     /// Obtain the n-th argument passed to the filter, crash if it is not given.
     // This function returns an `impl` in order not to expose the `Filter` type publicly.
     // It would be more elegant to implement `Index<usize>` here instead,
     // but because of returning `impl`, we cannot do this right now, see:
     // <https://github.com/rust-lang/rust/issues/63063>.
-    pub fn get(self, i: usize) -> impl FilterT<'a> {
+    // TODO for v2.0: make this take `&self`?
+    pub fn get(self, i: usize) -> impl FilterT<'a, V> {
         Ref(self.0[i], self.1)
     }
 }
 
-impl<'a> FilterT<'a> for &'a Owned {
-    fn run(self, cv: Cv<'a>) -> ValRs<'a> {
+impl<'a, V: ValT> FilterT<'a, V> for &'a Owned<V> {
+    fn run(self, cv: Cv<'a, V>) -> ValR2s<'a, V> {
         Ref(self.0, &self.1).run(cv)
     }
 
-    fn update(self, cv: Cv<'a>, f: Box<dyn Update<'a> + 'a>) -> ValRs<'a> {
+    fn update(self, cv: Cv<'a, V>, f: Box<dyn Update<'a, V> + 'a>) -> ValR2s<'a, V> {
         Ref(self.0, &self.1).update(cv, f)
     }
 }
 
-impl<'a> FilterT<'a> for Ref<'a> {
-    fn run(self, cv: Cv<'a>) -> ValRs<'a> {
+impl<'a, V: ValT> FilterT<'a, V> for Ref<'a, V> {
+    fn run(self, cv: Cv<'a, V>) -> ValR2s<'a, V> {
+        use alloc::string::ToString;
         use core::iter::{once, once_with};
         // wrap a filter AST with the filter definitions
         let w = move |id: &Id| Ref(*id, self.1);
-        match &self.1[self.0 .0] {
+        match &self.1.defs[self.0 .0] {
             Ast::Id => box_once(Ok(cv.1)),
-            Ast::ToString => Box::new(once_with(move || Ok(Val::str(cv.1.to_string_or_clone())))),
-            Ast::Int(n) => box_once(Ok(Val::Int(*n))),
-            Ast::Float(x) => box_once(Ok(Val::Float(*x))),
-            Ast::Str(s) => Box::new(once_with(move || Ok(Val::str(s.clone())))),
-            Ast::Array(f) => Box::new(once_with(move || {
-                w(f).run(cv).collect::<Result<_, _>>().map(Val::arr)
+            Ast::ToString => Box::new(once_with(move || match cv.1.as_str() {
+                Some(_) => Ok(cv.1),
+                None => Ok(V::from(cv.1.to_string())),
             })),
-            Ast::Object(o) if o.is_empty() => box_once(Ok(Val::Obj(Default::default()))),
-            Ast::Object(o) => Box::new(
-                obj_cart(o.iter().map(move |(k, v)| (w(k), w(v))), cv, Vec::new()).map(|kvs| {
-                    kvs.into_iter()
-                        .map(|(k, v)| Ok((k?.to_str()?, v?)))
-                        .collect::<Result<_, _>>()
-                        .map(Val::obj)
-                }),
-            ),
+            Ast::Int(n) => box_once(Ok(V::from(*n))),
+            Ast::Num(x) => box_once(V::from_num(x)),
+            Ast::Str(s) => Box::new(once_with(move || Ok(V::from(s.clone())))),
+            Ast::Array(f) => Box::new(once_with(move || w(f).run(cv).collect::<Result<_, _>>())),
+            Ast::ObjEmpty => box_once(V::from_map([])),
+            Ast::ObjSingle(k, v) => {
+                Box::new(Self::cartesian(w(k), w(v), cv).map(|(k, v)| V::from_map([(k?, v?)])))
+            }
             Ast::Try(f, c) => Box::new(w(f).run((cv.0.clone(), cv.1)).flat_map(move |y| {
                 y.map_or_else(
                     |e| w(c).run((cv.0.clone(), e.as_val())),
@@ -299,18 +307,19 @@ impl<'a> FilterT<'a> for Ref<'a> {
             Ast::Assign(path, f) => w(f).pipe(cv, move |cv, y| {
                 w(path).update(cv, Box::new(move |_| box_once(Ok(y.clone()))))
             }),
+
             Ast::Logic(l, stop, r) => w(l).pipe(cv, move |cv, l| {
                 if l.as_bool() == *stop {
-                    box_once(Ok(Val::Bool(*stop)))
+                    box_once(Ok(V::from(*stop)))
                 } else {
-                    Box::new(w(r).run(cv).map(|r| Ok(Val::Bool(r?.as_bool()))))
+                    Box::new(w(r).run(cv).map(|r| Ok(V::from(r?.as_bool()))))
                 }
             }),
             Ast::Math(l, op, r) => {
                 Box::new(Self::cartesian(w(l), w(r), cv).map(|(x, y)| op.run(x?, y?)))
             }
             Ast::Ord(l, op, r) => Box::new(
-                Self::cartesian(w(l), w(r), cv).map(|(x, y)| Ok(Val::Bool(op.run(&x?, &y?)))),
+                Self::cartesian(w(l), w(r), cv).map(|(x, y)| Ok(V::from(op.run(&x?, &y?)))),
             ),
 
             Ast::Fold(typ, xs, init, f) => {
@@ -337,14 +346,14 @@ impl<'a> FilterT<'a> for Ref<'a> {
                 let inputs = cv.0.inputs;
                 let cvs = bind_vars(call.args.iter().map(move |a| a.as_ref().map(w)), ctx, cv);
                 match call.typ {
-                    CallTyp::Normal => Box::new(run_cvs(def, cvs)),
+                    CallTyp::Normal => run_cvs(def, cvs),
                     CallTyp::Catch => Box::new(crate::Stack::new(
-                        Vec::from([Box::new(run_cvs(def, cvs)) as Results<_, _>]),
+                        Vec::from([run_cvs(def, cvs)]),
                         move |r| match r {
                             Err(Error::TailCall(TailCall(id, vars, v))) if id == call.id => {
-                                Err(def.run((Ctx { inputs, vars }, v)))
+                                ControlFlow::Continue(def.run((Ctx { inputs, vars }, v)))
                             }
-                            Ok(_) | Err(_) => Ok(r),
+                            Ok(_) | Err(_) => ControlFlow::Break(r),
                         },
                     )),
                     CallTyp::Throw => Box::new(cvs.map(move |cv| {
@@ -353,23 +362,24 @@ impl<'a> FilterT<'a> for Ref<'a> {
                 }
             }
 
-            Ast::Native(Native { run, .. }, args) => (run)(Args(args, self.1), cv),
+            Ast::Native(id, args) => (self.1.natives[*id].run)(Args(args, self.1), cv),
         }
     }
 
-    fn update(self, cv: Cv<'a>, f: Box<dyn Update<'a> + 'a>) -> ValRs<'a> {
+    fn update(self, cv: Cv<'a, V>, f: Box<dyn Update<'a, V> + 'a>) -> ValR2s<'a, V> {
         let err = box_once(Err(Error::PathExp));
         let w = move |id: &Id| Ref(*id, self.1);
-        match &self.1[self.0 .0] {
+        match &self.1.defs[self.0 .0] {
             Ast::ToString => err,
-            Ast::Int(_) | Ast::Float(_) | Ast::Str(_) => err,
-            Ast::Array(_) | Ast::Object(_) => err,
+            Ast::Int(_) | Ast::Num(_) | Ast::Str(_) => err,
+            Ast::Array(_) | Ast::ObjEmpty | Ast::ObjSingle(..) => err,
             Ast::Neg(_) | Ast::Logic(..) | Ast::Math(..) | Ast::Ord(..) => err,
             Ast::Update(..) | Ast::UpdateMath(..) | Ast::Assign(..) => err,
 
             // these are up for grabs to implement :)
-            Ast::Try(..) | Ast::Alt(..) => todo!(),
-            Ast::Fold(..) => todo!(),
+            Ast::Try(..) | Ast::Alt(..) | Ast::Fold(..) => {
+                unimplemented!("updating with this operator is not supported yet")
+            }
 
             Ast::Id => f(cv.1),
             Ast::Path(l, path) => {
@@ -412,40 +422,43 @@ impl<'a> FilterT<'a> for Ref<'a> {
                 reduce(cvs, init, move |cv, v| def.update((cv.0, v), f.clone()))
             }
 
-            Ast::Native(Native { update, .. }, args) => (update)(Args(args, self.1), cv, f),
+            Ast::Native(id, args) => (self.1.natives[*id].update)(Args(args, self.1), cv, f),
         }
     }
 }
 
+type Triple<T> = (T, T, T);
+
 /// Function from a value to a stream of value results.
 // TODO for v2.0: Remove `Clone` bound and make sub-trait requiring `Copy` for derived functions
-pub trait FilterT<'a>: Clone + 'a {
+// try to implement functions on &self?
+pub trait FilterT<'a, V: ValT = Val>: Clone + 'a {
     /// `f.run((c, v))` returns the output of `v | f` in the context `c`.
-    fn run(self, cv: Cv<'a>) -> ValRs<'a>;
+    fn run(self, cv: Cv<'a, V>) -> ValR2s<'a, V>;
 
     /// `p.update((c, v), f)` returns the output of `v | p |= f` in the context `c`.
-    fn update(self, cv: Cv<'a>, f: Box<dyn Update<'a> + 'a>) -> ValRs<'a>;
+    fn update(self, cv: Cv<'a, V>, f: Box<dyn Update<'a, V> + 'a>) -> ValR2s<'a, V>;
 
     /// For every value `v` returned by `self.run(cv)`, call `f(cv, v)` and return all results.
     ///
     /// This has a special optimisation for the case where only a single `v` is returned.
     /// In that case, we can consume `cv` instead of cloning it.
-    fn pipe<T: 'a, F>(self, cv: Cv<'a>, f: F) -> Results<'a, T, Error>
+    fn pipe<T: 'a, F>(self, cv: Cv<'a, V>, f: F) -> Results<'a, T, Error<V>>
     where
-        F: Fn(Cv<'a>, Val) -> Results<'a, T, Error> + 'a,
+        F: Fn(Cv<'a, V>, V) -> Results<'a, T, Error<V>> + 'a,
     {
         flat_map_with(self.run(cv.clone()), cv, move |y, cv| then(y, |y| f(cv, y)))
     }
 
     /// Run `self` and `r` and return the cartesian product of their outputs.
-    fn cartesian(self, r: Self, cv: Cv<'a>) -> BoxIter<'a, (ValR, ValR)> {
+    fn cartesian(self, r: Self, cv: Cv<'a, V>) -> BoxIter<'a, (ValR2<V>, ValR2<V>)> {
         flat_map_with(self.run(cv.clone()), cv, move |l, cv| {
             map_with(r.clone().run(cv), l, |r, l| (l, r))
         })
     }
 
     /// Run `self`, `m`, and `r`, and return the cartesian product of their outputs.
-    fn cartesian3(self, m: Self, r: Self, cv: Cv<'a>) -> BoxIter<'a, (ValR, ValR, ValR)> {
+    fn cartesian3(self, m: Self, r: Self, cv: Cv<'a, V>) -> BoxIter<'a, Triple<ValR2<V>>> {
         flat_map_with(self.run(cv.clone()), cv, move |l, cv| {
             let r = r.clone();
             flat_map_with(m.clone().run(cv.clone()), (l, cv), move |m, (l, cv)| {
@@ -461,14 +474,14 @@ pub trait FilterT<'a>: Clone + 'a {
     /// if `outer` is true, it returns values for which `f` yields no output.
     /// This is useful to implement `while` and `until`.
     #[deprecated(since = "1.2.0")]
-    fn recurse(self, inner: bool, outer: bool, cv: Cv<'a>) -> ValRs {
+    fn recurse(self, inner: bool, outer: bool, cv: Cv<'a, V>) -> ValR2s<V> {
         let f = move |v| self.clone().run((cv.0.clone(), v));
         Box::new(recurse(inner, outer, box_once(Ok(cv.1)), f))
     }
 
     /// Return the output of `recurse(l) |= f`.
     #[deprecated(since = "1.2.0")]
-    fn recurse_update(self, cv: Cv<'a>, f: Box<dyn Update<'a> + 'a>) -> ValRs<'a> {
+    fn recurse_update(self, cv: Cv<'a, V>, f: Box<dyn Update<'a, V> + 'a>) -> ValR2s<'a, V> {
         // implemented by the expansion of `def recurse(l): ., (l | recurse(l))`
         Box::new(f(cv.1).flat_map(move |v| {
             then(v, |v| {
