@@ -43,6 +43,8 @@ fn sub() {
     give(json!(1.0), ". - 1", json!(0.0));
 }
 
+yields!(sub_arr, "[1, 2, 3] - [2, 3, 4]", json!([1]));
+
 #[test]
 fn mul() {
     give(json!(1), ". * 2", json!(2));
@@ -61,6 +63,11 @@ fn mul() {
     );
 }
 
+yields!(div_str, r#""abcabcdab" / "ab""#, ["", "c", "cd", ""]);
+yields!(div_str_empty, r#""" / """#, json!([]));
+yields!(div_str_empty_str, r#""" / "ab""#, json!([]));
+yields!(div_str_empty_sep, r#""aöß" / """#, ["a", "ö", "ß"]);
+
 #[test]
 fn logic() {
     let tf = json!([true, false]);
