@@ -55,6 +55,14 @@ struct Cli {
     #[arg(short, long)]
     compact_output: bool,
 
+    /// Use n spaces for indentation
+    #[arg(long, value_name = "n", default_value_t = 2)]
+    indent: usize,
+
+    /// Use tabs for indentation rather than spaces
+    #[arg(long)]
+    tab: bool,
+
     /// Do not print a newline after each value
     ///
     /// Unlike jq, this does not enable `--raw-output`.
@@ -153,7 +161,7 @@ fn real_main() -> Result<ExitCode, Error> {
         let mut last = None;
         for file in files {
             let path = std::path::Path::new(file);
-            let file = mmap_file(path).map_err(|e| Error::Io(Some(file.to_string()), e))?;
+            let file = load_file(path).map_err(|e| Error::Io(Some(file.to_string()), e))?;
             let inputs = read_slice(&cli, &file);
             if cli.in_place {
                 // create a temporary file where output is written to
@@ -213,7 +221,7 @@ fn binds(cli: &Cli) -> Result<Vec<(String, Val)>, Error> {
     })?;
     bind(&mut var_val, &cli.slurpfile, |f| {
         let path = std::path::Path::new(f);
-        let file = mmap_file(path).map_err(|e| Error::Io(Some(f.to_string()), e))?;
+        let file = load_file(path).map_err(|e| Error::Io(Some(f.to_string()), e))?;
         Ok(Val::arr(json_slice(&file).collect::<Result<Vec<_>, _>>()?))
     })?;
 
@@ -262,9 +270,13 @@ fn parse(filter_str: &str, vars: Vec<String>) -> Result<Filter, Vec<ParseError>>
     }
 }
 
-fn mmap_file(path: &std::path::Path) -> io::Result<memmap2::Mmap> {
+/// Try to load file by memory mapping and fall back to regular loading if it fails.
+fn load_file(path: &std::path::Path) -> io::Result<Box<dyn core::ops::Deref<Target = [u8]>>> {
     let file = std::fs::File::open(path)?;
-    unsafe { memmap2::Mmap::map(&file) }
+    match unsafe { memmap2::Mmap::map(&file) } {
+        Ok(mmap) => Ok(Box::new(mmap)),
+        Err(_) => Ok(Box::new(std::fs::read(path)?)),
+    }
 }
 
 fn invalid_data(e: impl std::error::Error + Send + Sync + 'static) -> std::io::Error {
@@ -403,6 +415,10 @@ impl From<Vec<ParseError>> for Error {
     }
 }
 
+/// Run a filter with given input values and run `f` for every value output.
+///
+/// This function cannot return an `Iterator` because it creates an `RcIter`.
+/// This is most unfortunate. We should think about how to simplify this ...
 fn run(
     cli: &Cli,
     filter: &Filter,
@@ -440,6 +456,11 @@ fn print(cli: &Cli, val: Val, writer: &mut impl Write) -> io::Result<()> {
         _ => {
             let val = serde_json::Value::from(val);
             let mode = cli.color_mode();
+            let indent = if cli.tab {
+                String::from("\t")
+            } else {
+                " ".repeat(cli.indent)
+            };
 
             // this looks ugly, but it is hard to abstract over the `Formatter` because
             // we cannot create a `Box<dyn Formatter>` because
@@ -447,7 +468,8 @@ fn print(cli: &Cli, val: Val, writer: &mut impl Write) -> io::Result<()> {
             if cli.compact_output {
                 ColoredFormatter::new(CompactFormatter).write_colored_json(&val, writer, mode)
             } else {
-                ColoredFormatter::new(PrettyFormatter::new()).write_colored_json(&val, writer, mode)
+                ColoredFormatter::new(PrettyFormatter::with_indent(indent.as_bytes()))
+                    .write_colored_json(&val, writer, mode)
             }?;
         }
     };
